/*  DIASTEXT.C is a translation of my Turbo Pascal program of the same
	name.  It automates a version of Jackson Mac Low's "diastic" method
	of manipulating	text.  (See Mac Low's THE VIRGINIA WOOLF POEMS, 
	Burning Deck Press,	for explanation.)  The main change to his method
	is that the text itself, rather than simply providing the key (as 
	in his "Ridiculous in Piccadilly") itself as a whole becomes the key.

	This version expects an input file to be named on the command line.
	If a second filename is given, it is used for output (diastic text
	only, not initial and terminal messages), though output is sent to
	the screen in any case.	

	Charles O. Hartman
	129 Wentworth Ave., Cranston, RI 02905
	(401) 461-3165
	QuickC version March 1989
*/

#include <stdio.h>
#include <ctype.h>
#include <memory.h>
#include <time.h>

#define MAXCHARS		32 * 1024
#define MAXWORDCHARS    30
#define MAXLINELEN      50
#define TABSIZE         8
#define TRUE            1
#define FALSE           0

#define err_exit(s)     (fputs((s),stderr), exit(1))
#define iswordchar(c)   (isalnum(c) || (c) == '-' || (c) == '\'')
#define isterminalpunct(c)  ((c) == '.' || (c) == '?' || (c) == '!')

typedef int bool;

char text_array[MAXCHARS];  /* buffer holding whole text to be treated */
char *text;         /* pointer to start (adjusted) of text in buffer */
int text_chars;		/* total number of characters read from input */
int re_entrances = 0;	/* times the text is re-used for search */
int line_pos = 0;   /* position on output line */
bool last_not_punct = TRUE;
bool file_output = FALSE;
char *key;          /* pointer into text array for current key character */
/* Note: the "key" is the character/position for which the next match is
   being sought.  It moves steadily through the input text, skipping all
   whitespace and punctuation, until the whole text has been used as a
   key.  Fully characterizing the current key requires both the character
   pointed to by (char *)key and the position-within-a-word indexed by
   word_pos; so the comments speak of both a "key character" and a
   "key position"--the word_pos that follows.  */
int word_pos = 0;       /* position-in-word of current key */
char *next_word;    /* points to start of potential next output word */
FILE *outf;

/*      function prototypes     */
bool set_key(void);
bool find_match(void);
void write_out(char *word_pointer);
void new_line(void);

main(int argc, char **argv)
{
	FILE *inf;
	time_t start, stop;

	if (argc < 2) err_exit("Must supply name of input file");
	if ((inf = fopen(argv[1], "rt")) == NULL)
		err_exit("Can't open input file");
	text = text_array;
	if ((text_chars = fread(text, 1, MAXCHARS, inf)) >= MAXCHARS-2) {
		fclose(inf);
		err_exit("Input text too long - 32k limit");
	}
	fclose(inf);
	if (text_chars < 1) err_exit("No input text received.");
	if (argc > 2) {
		file_output = TRUE;
		if ((outf = fopen(argv[2], "wt")) == NULL)
			err_exit("Can't open output file");
	}
	time(&start);
	while (!isalpha(*text)) {
		text++;
		text_chars--;
	}
	write_out(text);    /* prime with first word */
	key = text;
	while (!set_key());
	next_word = key;
	while ((key - text) < text_chars) {
		while(!find_match());
		while(!set_key());
	}
	new_line();
	time(&stop);
 	if (file_output) fclose(outf);
	fprintf(stderr, "\n\nSTATS:\n%d characters read\n", text_chars);
	fprintf(stderr, "%d re-entrances\n", re_entrances);
	fprintf(stderr, "%d seconds\n", stop-start);
 }   /* main() */

bool set_key(void)
{
	if (kbhit()) err_exit("\n\nKey struck - processing halted");
	if ((++key - text) >= text_chars) return(TRUE); /* and get out */
	if (iswordchar(*key)) {
		word_pos++;
		if (isalpha(*key)) return(TRUE);
	}
	else word_pos = -1;
	return(FALSE); 		/* no new key */
}   /* set_key() */

bool find_match(void)
{
	int i = 0;
	char *cp;

	if (kbhit()) err_exit("\n\nKey struck - processing halted");
	while (iswordchar(*next_word)) ++next_word;
	while (!iswordchar(*next_word)) {
		if ((next_word - text) >= text_chars) {
			re_entrances++;
			next_word = text;
		}
		else ++next_word;
	}
	for (cp = next_word, i = 0; i < word_pos; cp++, i++)
		if (!iswordchar(*cp)) return(FALSE);    /* still in word? */
	if (tolower(*cp) == tolower(*key)) {      /* success */
		write_out(next_word);
		return(TRUE);
	}
	return(FALSE);
}   /* find_match() */

void write_out(char *wptr)
{
	char *cp;

	if (isupper(*wptr) && last_not_punct) new_line();
	for (cp = wptr; isgraph(*cp); cp++, line_pos++) {
		putc(*cp, stdout);
		if (file_output) putc(*cp, outf);
	}
	putc(' ', stdout);
	if (file_output) putc(' ', outf);
	if (ispunct(*(--cp))) {		/* back to look at last one */
		last_not_punct = FALSE;
		new_line();
		if (isterminalpunct(*cp)) new_line();
	}
	else last_not_punct = TRUE;
	if (line_pos >= MAXLINELEN) new_line();
}   /* write_out() */

void new_line(void)
{
	putc('\n', stdout);
	if (file_output) putc('\n', outf);
	if (line_pos >= MAXLINELEN) {
		fputs("        ", stdout);
		if (file_output) fputs("        ", outf);
		line_pos = TABSIZE;
	}
	else line_pos = 0;
}   /* new_line() */
